/* Place in the public domain by Sam Trenholme 2000 */
/* js_string routines that depend on the code page (encoding) that we
   are in */

#include "JessStr.h"

/* js_newline_chars: returns a set of newline characters for a given
                     encoding
   input: an empty js_string object we look at the codepage of
   output: JS_ERROR (-1) on error, JS_SUCCESS (1) on success */

int js_newline_chars(js_string *js) {
   
    /* Sanity check */
    if(js_has_sanity(js) == -1)
        return -1;
   
    /* Handle different encodings differently */ 
    switch(js->encoding) {
        case JS_BINARY:
            js->unit_count = 0;
            return JS_SUCCESS;
        case JS_US_ASCII:
            if(js->unit_size != 1)
                return -1; /* This string is not sane */ 
            if(js_octets(js) < 2)
                return -1; /* No overflows ever */
            *(js->string) = '\r';
            *(js->string + 1) = '\n';
            js->unit_count = 2;
            return JS_SUCCESS;
        }

     /* We only get here if it is an unhandled encoding */
     return JS_ERROR;

     }
 
/* js_space_chars: returns a set of whitespace characters for a given
                   encoding
   input: an empty js_string object we look at the codepage of
   output: JS_ERROR (-1) on error, JS_SUCCESS (1) on success */

int js_space_chars(js_string *js) {
   
    /* Sanity check */
    if(js_has_sanity(js) == -1)
        return -1;
   
    /* Handle different encodings differently */ 
    switch(js->encoding) {
        case JS_BINARY:
            js->unit_count = 0;
            return JS_SUCCESS;
        case JS_US_ASCII:
            if(js->unit_size != 1)
                return -1; /* This string is not sane */ 
            if(js_octets(js) < 2)
                return -1; /* No overflows ever */
            *(js->string) = ' ';
            *(js->string + 1) = '\t';
            js->unit_count = 2;
            return JS_SUCCESS;
        }

     /* We only get here if it is an unhandled encoding */
     return JS_ERROR;

     }
 
/* js_atoi: Convert a number, starting at a given offset, into an integer 
   input: js_string object to look at, place in string to start looking at
   output: integer we find at the given point, 0 on error (ugh) */
int js_atoi(js_string *js, int offset) {
   
   int value, sign;
 
   if(js_has_sanity(js) == JS_ERROR)
       return 0;

   if(js->encoding != JS_US_ASCII) /* Only US ASCII support right now */
       return 0; 
   else {
       if(offset >= js->unit_count)
           return 0;
       value = 0;
       sign = 1;
       while(offset < js->unit_count && *(js->string + offset) >= '0' 
             && *(js->string + offset) <= '9') {
           value *= 10;
           value += *(js->string + offset) - '0';
           offset++;
           } 
       value *= sign;
       return value; 
       }

   return 0; /* We should never get here */ 
   }

/* js_tolower: Convert a js_string object in to all lower case letters.
   input: js_string object to convert
   output: JS_ERROR on error, JS_SUCCESS on success */
int js_tolower(js_string *js) {
    int counter;

    if(js_has_sanity(js) == JS_ERROR)
        return JS_ERROR;

    if(js->unit_size != 1)
        return JS_ERROR;

    if(js->encoding != JS_US_ASCII) /* Only US ASCII support right now */
        return 0;
    else 
        for(counter = 0; counter < js->unit_count; counter++)
	    if(*(js->string + counter) >= 'A' && 
	       *(js->string + counter) <= 'Z')
	        *(js->string + counter) += 32;
 
    return JS_SUCCESS;
    }

